#include <ArduinoBLE.h>
#include <Arduino_LSM9DS1.h>
#include <Wire.h>
#include <VL53L0X.h>
#include <Adafruit_NeoPixel.h>
// Declarations
// Time cycle
unsigned long currentMillis = 0, previousMillis = 0;
long const tickLength = 13;
int cycleCount = 0;
// BLE
byte midiData[] = {0x80, 0x80, 0x00, 0x00, 0x00};
byte progData[] = {0x80, 0x80, 0x00, 0x00};
char const DEVICE_NAME[] = "Xeramin_R1";
BLEService midiService("03B80E5A-EDE8-4B33-A751-6CE34EC4C700");
BLECharacteristic midiCharacteristic("7772E5DB-3868-4112-A1A9-F2669D106BF3",
                                     BLEWriteWithoutResponse | BLENotify | BLERead, sizeof(midiData));
bool centralConnected = false;
// MechKey
#define BUTTON_PIN 2
#define PIXEL_PIN 3
#define PIXEL_COUNT 1
Adafruit_NeoPixel keyLED = Adafruit_NeoPixel(PIXEL_COUNT, PIXEL_PIN, NEO_GRB);
byte const colorBank[] = {255, 0, 0, 127, 127, 0, 0, 255, 0, 0, 127, 127, 0, 0, 255, 127, 0, 127, 0, 0, 0};
int colorCode = 6;
bool keyState[] = {false, false};
int kIdx = 0;
bool keyPressed = false;
// IMU data
VL53L0X sensor;
int aIdx = 0;
float x[] = {0, 0, 0}, y[] = {0, 0, 0}, z[] = {0, 0, 0};
int const SIZEOF_ACC_ARR = 3;
float maxMinAcc[] = {0, 0};
// Acc thresholds keys and constants
#define NEW_NOTE 0
#define SUSTAIN_NOTE 1
#define REPEAT_NOTE 2
float const Z_GATING_RATIO = 0.25;
float const Z_GATING_RATIO_2 = 0.28;
float const DEFAULT_X_THRESHOLDS[] = {0.40, 0.40, 0.60}, DEFAULT_Y_THRESHOLDS[] = {0.35, 0.35, 0.55};
float xThresholds[] = {0.40, 0.40, 0.60}, yThresholds[] = {0.35, 0.35, 0.55};
// Laser range data
#define LASER_SRL 0.25
#define LASER_TIMEBUDGET 25000
int dist[] = {0, 0, 0};
int dIdx = 0;
// MIDI channels and soundings
int cIdx = -1;
int const SIZEOF_CH_ARR = 3;
byte const ch[] = {1, 2, 3, 10};
byte noteInPlay[] = {0, 0, 0, 0};
int noteLife[] = {0, 0, 0, 0};
int silenceTicks = 0;
bool playFlat = false, anyNotePlayed = false;
int const DEFAULT_LIFE = 75, DEFAULT_SUSTAIN = 40, DEFAULT_DECAY = 25;
// Note setting
byte noteToPlay = 0;
byte velToPlay = 0;
byte const notes[] = {
  0,
  53, 55, 57, 59,
  60, 62, 64, 65, 67, 69, 71,
  72, 74, 76, 77, 79, 81, 83,
  84, 86, 88
};
int const AVAIL_NOTE_STEPS = 21;
int const NOTE_STEP_DIST = 88;
// Volume and touches
float const VOL_THRESHOLD[] = {0.75, 0.45}, VEL_THRESHOLD[] = {1.10, 0.80};
int const VOL_TABLE[] = {120, 100, 80}, VEL_TABLE[] = {120, 100, 80};
int const SIZEOF_VOL_TABLE = 3, SIZEOF_VEL_TABLE = 3;
byte const DEFAULT_VOL = 100, DEFAULT_VEL = 100;

// Main program
void setup() {
  setupIMU();
  setupBLE();
  setupLaser();
  setupKey();
}
void loop() {
  currentMillis = millis();
  if (currentMillis - previousMillis >= tickLength) {
    previousMillis = currentMillis;
    cycleCount = (cycleCount + 1) % 4;
    switch (cycleCount) {
      case 0:
        updateAcc();
        updateKeyState();
        makeSound();
        break;
      case 1:
        updateRange();
        updateAcc();
        updateKeyState();
        break;
      case 2:
        updateLED();
        updateAcc();
        updateKeyState();
        updateSound();
        break;
      case 3:
        updateRange();
        updateAcc();
        updateKeyState();
        break;
    }
  }
}

// key and LED
void setupKey() {
  pinMode(BUTTON_PIN, INPUT_PULLUP);
}

void updateKeyState() {
  keyState[kIdx] = (digitalRead(BUTTON_PIN) == HIGH) ? true : false;
  kIdx = kIdx ? 0 : 1;
  if (keyState[0] && keyState[1]) keyPressed = true;
  if (!keyState[0] && !keyState[1]) keyPressed = false;
}

void updateLED() {
  int mDist = median(dist);
  noteToPlay = distToNote(mDist);
  if (noteToPlay && centralConnected) {
    colorCode = noteToPlay % 6;
  } else {
    colorCode = 6;
  }
  keyLED.setPixelColor(0, code2Color(colorCode));
  keyLED.show();
}

uint32_t code2Color(int code) {
  return keyLED.Color(colorBank[(code * 3)], colorBank[(code * 3 + 1)], colorBank[(code * 3 + 2)]);
}

// state 0 machine
void makeSound() {
  BLEDevice central = BLE.central();
  int mDist = median(dist);
  noteToPlay = distToNote(mDist);
  maxMin(maxMinAcc, z, SIZEOF_ACC_ARR);
  float zAccGate = (maxMinAcc[0] - maxMinAcc[1]) * Z_GATING_RATIO;
  maxMin(maxMinAcc, y, SIZEOF_ACC_ARR);
  float yAccMax = max(abs(maxMinAcc[0]), abs(maxMinAcc[1])) - zAccGate;
  float yAccSpan = maxMinAcc[0] - maxMinAcc[1];
  maxMin(maxMinAcc, x, SIZEOF_ACC_ARR);
  float xAccMax = max(abs(maxMinAcc[0]), abs(maxMinAcc[1])) - zAccGate;
  float xAccSpan = maxMinAcc[0] - maxMinAcc[1];
  if (xAccMax > yAccMax) {
    playFlat = true;
    noteToPlay = noteToPlay ? noteToPlay - 1 : 0;
  } else {
    playFlat = false;
  }
  bool goodToGo = ((!keyPressed) && noteToPlay && central) ? true : false;
  if (goodToGo && playFlat) {
    cIdx = lookupNoteArrays(noteInPlay, noteLife, noteToPlay, SIZEOF_CH_ARR);
    if (cIdx == -1) {
      // new flat Note
      if (xAccMax > xThresholds[NEW_NOTE]) {
        velToPlay = accToVel(xAccSpan, SIZEOF_VEL_TABLE);
        replaceNote();
        xThresholds[SUSTAIN_NOTE] = 0;
        xThresholds[REPEAT_NOTE] = DEFAULT_X_THRESHOLDS[REPEAT_NOTE];
        anyNotePlayed = true;
      }
    } else if (xAccMax > xThresholds[REPEAT_NOTE]) {
      // repeat same flat Note
      velToPlay = accToVel(xAccSpan, SIZEOF_VEL_TABLE);
      replaceNote();
      xThresholds[REPEAT_NOTE] = DEFAULT_X_THRESHOLDS[REPEAT_NOTE];
      anyNotePlayed = true;
    } else if (xAccMax > xThresholds[SUSTAIN_NOTE]) {
      // sustain flat Note
      noteLife[cIdx] += DEFAULT_SUSTAIN;
      xThresholds[SUSTAIN_NOTE] = (xThresholds[SUSTAIN_NOTE] + DEFAULT_X_THRESHOLDS[SUSTAIN_NOTE]) / 2;
      xThresholds[REPEAT_NOTE] = (xAccMax + DEFAULT_X_THRESHOLDS[REPEAT_NOTE]) / 2;
      anyNotePlayed = true;
    }
  } else if (goodToGo) {
    cIdx = lookupNoteArrays(noteInPlay, noteLife, noteToPlay, SIZEOF_CH_ARR);
    if (cIdx == -1) {
      // new Note
      if (yAccMax > yThresholds[NEW_NOTE]) {
        velToPlay = accToVel(yAccSpan, SIZEOF_VEL_TABLE);
        replaceNote();
        yThresholds[SUSTAIN_NOTE] = 0;
        yThresholds[REPEAT_NOTE] = DEFAULT_Y_THRESHOLDS[REPEAT_NOTE];
        anyNotePlayed = true;
      }
    } else if (yAccMax > yThresholds[REPEAT_NOTE]) {
      // repeat same Note
      velToPlay = accToVel(yAccSpan, SIZEOF_VEL_TABLE);
      replaceNote();
      yThresholds[REPEAT_NOTE] = DEFAULT_Y_THRESHOLDS[REPEAT_NOTE];
      anyNotePlayed = true;
    } else if (yAccMax > yThresholds[SUSTAIN_NOTE]) {
      // sustain Note
      noteLife[cIdx] += DEFAULT_SUSTAIN;
      yThresholds[SUSTAIN_NOTE] = (yThresholds[SUSTAIN_NOTE] + DEFAULT_Y_THRESHOLDS[SUSTAIN_NOTE]) / 2;
      yThresholds[REPEAT_NOTE] = (yAccMax + DEFAULT_Y_THRESHOLDS[REPEAT_NOTE]) / 2;
      anyNotePlayed = true;
    }
  }
  if ((anyNotePlayed && yAccMax < yThresholds[SUSTAIN_NOTE] && xAccMax < xThresholds[SUSTAIN_NOTE])
      || (keyPressed) ) {
    silenceTicks++;
    if (silenceTicks > DEFAULT_LIFE) {
      for (int i = 0; i < SIZEOF_CH_ARR; i++) allOff(ch[i]);
      silenceTicks = 0;
      anyNotePlayed = false;
    }
  }
}

void replaceNote() {
  cIdx = lookupLeast(noteLife, SIZEOF_CH_ARR);
  if (noteInPlay[cIdx]) noteOff(ch[cIdx], noteInPlay[cIdx]);
  setVol(ch[cIdx], DEFAULT_VOL);
  noteOn(ch[cIdx], noteToPlay, velToPlay);
  noteInPlay[cIdx] = noteToPlay;
  noteLife[cIdx] = DEFAULT_LIFE;
}

// state 2 machine
void updateSound() {
  maxMin(maxMinAcc, z, SIZEOF_ACC_ARR);
  float zAccGate = (maxMinAcc[0] - maxMinAcc[1]) * Z_GATING_RATIO_2;
  maxMin(maxMinAcc, y, SIZEOF_ACC_ARR);
  float yAccSpan = maxMinAcc[0] - maxMinAcc[1] - zAccGate;
  maxMin(maxMinAcc, x, SIZEOF_ACC_ARR);
  float xAccSpan = maxMinAcc[0] - maxMinAcc[1] - zAccGate;
  byte vol = VOL_TABLE[(SIZEOF_VOL_TABLE - 1)];
  for (int i = 0; i < SIZEOF_CH_ARR; i++) {
    int life = noteLife[i] - DEFAULT_DECAY;
    if (noteLife[i] > 0 && life <= 0) {
      // NoteOff decayed notes
      noteOff(ch[i], noteInPlay[i]);
      setVol(ch[i], DEFAULT_VOL);
      noteInPlay[i] = 0;
      noteLife[i] = 0;
    } else {
      // Adjust volume
      if (noteToPlay && noteToPlay == noteInPlay[i]) {
        if (!keyPressed) vol = accToVol(xAccSpan, yAccSpan, SIZEOF_VOL_TABLE);
        setVol(ch[i], vol);
      }
      noteLife[i] = max(0, life);
    }
  }
}

// Data handling and conversion
byte distToNote(int distance) {
  int indexTaken = distance / NOTE_STEP_DIST;
  byte legalNote = (indexTaken > 0 && indexTaken < AVAIL_NOTE_STEPS) ? notes[indexTaken] : 0;
  return legalNote;
}

byte accToVel(float acc, int arrSize) {
  arrSize--;
  byte vel = VEL_TABLE[arrSize];
  while (arrSize > 0) {
    arrSize--;
    if (acc > VEL_THRESHOLD[arrSize]) vel = VEL_TABLE[arrSize];
  }
  return vel;
}

byte accToVol(float xAcc, float yAcc, int arrSize) {
  xAcc = (yAcc > xAcc) ? yAcc : xAcc;
  arrSize--;
  byte vol = VOL_TABLE[arrSize];
  while (arrSize > 0) {
    arrSize--;
    if (xAcc > VOL_THRESHOLD[arrSize]) vol = VOL_TABLE[arrSize];
  }
  return vol;
}

void updateAcc() {
  if (IMU.accelerationAvailable()) {
    IMU.readAcceleration(x[aIdx], y[aIdx], z[aIdx]);
    aIdx = (aIdx + 1) % 3;
  }
}

void updateRange() {
  int reading = sensor.readRangeContinuousMillimeters();
  dist[dIdx] = reading;
  dIdx = (dIdx + 1) % 3;
}

int lookupNoteArrays(byte arr[], int arr2[], byte item, int arrSize) {
  int idx = -1;
  int matchCount = 0;
  for (int i = 0; i < arrSize; i++) {
    if (item == arr[i]) {
      matchCount++;
      if (matchCount == 1) idx = i;
      if (matchCount > 1 && arr2[i] > arr2[idx]) idx = i;
    }
  }
  return idx;
}

// Math methods
int lookupLeast(int arr[], int arrSize) {
  int idx = 0;
  int lastVal = arr[0];
  for (int i = 1; i < arrSize; i++) {
    if (arr[i] < lastVal) {
      idx = i;
      lastVal = arr[i];
    }
  }
  return idx;
}

int median(int numbers[]) {
  if ((numbers[0] >= numbers[1] && numbers[0] < numbers[2]) || (numbers[0] < numbers[1] && numbers[0] >= numbers[2])) return numbers[0];
  if ((numbers[1] >= numbers[0] && numbers[1] < numbers[2]) || (numbers[1] < numbers[0] && numbers[1] >= numbers[2])) return numbers[1];
  return numbers[2];
}

void maxMin(float answer[], float arr[], int arrSize) {
  arrSize--;
  float maximum = arr[arrSize];
  float minimum = arr[arrSize];
  while (arrSize > 0) {
    arrSize--;
    maximum = max(maximum, arr[arrSize]);
    minimum = min(minimum, arr[arrSize]);
  }
  answer[0] = maximum;
  answer[1] = minimum;
}

// Setup hardwares
void setupLaser() {
  Wire.begin();
  sensor.setTimeout(500);
  if (!sensor.init())
  {
    while (true);
  }
  sensor.setSignalRateLimit(LASER_SRL);
  sensor.setMeasurementTimingBudget(LASER_TIMEBUDGET);
  sensor.startContinuous();
}

void setupIMU() {
  if (!IMU.begin()) {
    while (true);
  }
}

void setupBLE() {
  if (!BLE.begin()) {
    while (true);
  }
  BLE.setLocalName(DEVICE_NAME);
  BLE.setDeviceName(DEVICE_NAME);
  BLE.setAdvertisedService(midiService);
  BLE.setEventHandler(BLEConnected, onConnected);
  BLE.setEventHandler(BLEDisconnected, onDisconnected);
  midiCharacteristic.setEventHandler(BLESubscribed, midiServiceSubscribed);
  midiService.addCharacteristic(midiCharacteristic);
  BLE.addService(midiService);
  BLE.setConnectable(true);
  BLE.setAdvertisingInterval(32);
  BLE.setConnectionInterval(9, 12);
  BLE.advertise();
}

void midiCommand(byte cmd, byte data1, byte data2) {
  midiData[2] = cmd;
  midiData[3] = data1;
  midiData[4] = data2;
  midiCharacteristic.setValue(midiData, sizeof(midiData));
}

void progCommand(byte cmd, byte data) {
  progData[2] = cmd;
  progData[3] = data;
  midiCharacteristic.setValue(progData, sizeof(progData));
}

void onConnected(BLEDevice central) {
  centralConnected = true;
  midiCharacteristic.setValue(0);
}

void onDisconnected(BLEDevice central) {
  centralConnected = false;
}

void midiServiceSubscribed(BLEDevice central, BLECharacteristic characteristic) {
}

// Midi commands intepretation
void noteOn(byte ch, byte note, byte vel) {
  byte chByte = ch + 143;
  midiCommand(chByte, note, vel);
}

void noteOff(byte ch, byte note) {
  byte chByte = ch + 127;
  midiCommand(chByte, note, 0);
}

void setVol(byte ch, byte vol) {
  byte chByte = ch + 175;
  midiCommand(chByte, 7, vol);
}

void allOff(byte ch) {
  byte chByte = ch + 175;
  midiCommand(chByte, 123, 0);
}
